#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
#
# Author:
# Emma E. M. Hobbs
#
# Contact
# eemh1@st-andrews.ac.uk
#
# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK
#
# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Tests scripts in the script module

These test are intened to be run from the root of the repository using:
pytest -v
"""


import logging
import pytest
import subprocess

from argparse import Namespace
from pathlib import Path

from saintBioutils.utilities import file_io

from cazomevolve.cazome.dbcan import invoke_dbcan



@pytest.fixture
def argsdict(test_input_dir, test_output_dir):
    out = test_output_dir / "run_dbcan"
    return {'args': Namespace(
        output_dir=out,
        force=False,
        nodelete=True,
        input_dir=test_input_dir,
        dbcan_version=None,
        cpu=8,
    )}


def test_run_dbcan_main(argsdict, monkeypatch):
    """Test the main entry point for build_cazy_db.py"""    
    def mock_run(*args, **kwards):
        return
    def mock_files(*args, **kwards):
        return [Path('GCA_012345678.1.faa'), Path('TEST'), Path('GCA_1111111111.1.faa')]
    
    monkeypatch.setattr(subprocess, "run", mock_run)
    monkeypatch.setattr(invoke_dbcan, "make_output_directory", mock_run)
    monkeypatch.setattr(invoke_dbcan, "get_file_paths", mock_files)
    monkeypatch.setattr(invoke_dbcan, "invoke_dbcan", mock_run)
    monkeypatch.setattr(invoke_dbcan, "closing_message", mock_run)

    invoke_dbcan.main(args=argsdict['args'], logger=logging.getLogger(__name__))



def test_run_dbcan_2(argsdict, monkeypatch):
    """Test running dbcan version 2"""    
    def mock_run(*args, **kwards):
        return
    
    argsdict['args'].dbcan_version = 2
    
    monkeypatch.setattr(subprocess, "run", mock_run)
    monkeypatch.setattr(invoke_dbcan, "make_output_directory", mock_run)

    invoke_dbcan.invoke_dbcan(
        Path('GCA_012345678.1.faa'),
        Path('tests/test_outputs/run_dbcan'),
        args=argsdict['args'],
    )


def test_run_dbcan_3(argsdict, monkeypatch):
    """Test running dbcan version 2"""    
    def mock_run(*args, **kwards):
        return
    
    argsdict['args'].dbcan_version = 3

    monkeypatch.setattr(subprocess, "run", mock_run)
    monkeypatch.setattr(invoke_dbcan, "make_output_directory", mock_run)

    invoke_dbcan.invoke_dbcan(
        Path('GCA_012345678.1.faa'),
        Path('tests/test_outputs/run_dbcan'),
        args=argsdict['args'],
    )



def test_run_dbcan_4(argsdict, monkeypatch):
    """Test running dbcan version 2"""    
    def mock_run(*args, **kwards):
        return
    
    argsdict['args'].dbcan_version = 4

    monkeypatch.setattr(subprocess, "run", mock_run)
    monkeypatch.setattr(invoke_dbcan, "make_output_directory", mock_run)

    invoke_dbcan.invoke_dbcan(
        Path('GCA_012345678.1.faa'),
        Path('tests/test_outputs/run_dbcan'),
        args=argsdict['args'],
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2020-2021
# (c) University of Strathclyde 2020-2021
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Retrieve all genomic assembly accessions descendent from a taxonomy node"""


import logging
import re
import time

from pathlib import Path
from typing import List, Optional

from socket import timeout
from Bio import Entrez
from saintBioutils.utilities.file_io import make_output_directory
from tqdm import tqdm
from urllib.request import urlopen
from urllib.error import HTTPError, URLError

from cazomevolve import closing_message


def main(args: Optional[List[str]] = None, logger: Optional[logging.Logger] = None):
    """Coordinate the retrieval of protein annotations from GenBank (.gbff) files.
    Including building parser, logger and output directory.
    Return dataframe of protein data.
    """
    make_output_directory(args.output_dir, args.force, args.nodelete)

    Entrez.email = args.email

    uid_lists = []

    for term in tqdm(((args.terms).split(",")), desc="Searching NCBI with terms"):
        uid_list = get_id_list(term)
        if uid_list is None:
            continue
        uid_lists.append(uid_list)

    for uid_list in tqdm(uid_lists, desc="Processing UID lists"):
        get_tax_ids(uid_list, term, args)

    closing_message('Download genomes', args)


def get_id_list(term):
    """Retrieve UIDs for all nodes below term.
    :param term: str, term to search NCBI Assembly with
    Return list of UIDs from NCBI Assembly database.
    """
    logger = logging.getLogger(__name__)

    with entrez_retry(
        Entrez.esearch,
        db="Assembly",
        term=term,
        idtype='acc',
        retmax=10000,
    ) as record_handle:
        record = Entrez.read(record_handle, validate=False)

    if len(record['IdList']) == 0:
        logger.warning(f"Retrieved 0 UIDs for {term}")
        return None

    return record['IdList']


def get_tax_ids(uid_list, term, args):
    """Retrieve Taxonomy ID for each UID and write to output file.

    :param uid_list: list of UIDs from NCBI
    :param term: str, term used to retrieve UIDs
    :param args: cmd-line args parser

    Return nothing.
    """
    logger = logging.getLogger(__name__)

    # batch query UIDs
    with entrez_retry(Entrez.epost, db="Assembly", id=(",".join(uid_list))) as record_handle:
        epost_result = Entrez.read(record_handle, validate=False)

    epost_webenv = epost_result["WebEnv"]
    epost_query_key = epost_result["QueryKey"]

    # retrieve summary document for each UID
    with entrez_retry(
        Entrez.efetch,
        db="Assembly",
        query_key=epost_query_key,
        WebEnv=epost_webenv,
        rettype="docsum",
        retmode="xml"
    ) as batch_handle:
        batch_result = Entrez.read(batch_handle, validate=False)

    accession_data = {}  # {Assembly accession : Assembly Name}

    for index in range(len(batch_result['DocumentSummarySet']['DocumentSummary'])):
        proceed = False
        if 'all' in args.assembly_levels:
            proceed = True
        else:
            assembly_level = batch_result['DocumentSummarySet']['DocumentSummary'][index]['AssemblyStatus'].lower()
            if assembly_level in args.assembly_levels:
                proceed = True

        if proceed is False:
            continue

        accession_number_v = batch_result['DocumentSummarySet']['DocumentSummary'][index]['AssemblyAccession']  # includes version number
        accession_number = accession_number_v[: (accession_number_v.find("."))]  # accession number excluding version number
        search_results = [acc for acc in list(accession_data.keys()) if acc.startswith(accession_number)]

        if len(search_results) == 0:  # no other versions of the assembly found
            accession_data[accession_number_v] = batch_result['DocumentSummarySet']['DocumentSummary'][index]['AssemblyName']
        
        elif len(search_results) == 1:
            accession_v = accession_number_v[(accession_number_v.find(".")) + 1 : ]
            existing_accession = search_results[0]
            existing_v = existing_accession[(existing_accession.find(".")) + 1 : ]

            if accession_v > existing_v:
                # the new accession number is a new version that the assembly already in the dict
                # remove the old assembly
                accession_data = accession_data.pop(existing_v)
                # add the newer version
                accession_data[accession_number_v] = batch_result['DocumentSummarySet']['DocumentSummary'][index]['AssemblyName']

            # else: existing version is newer so do no overwrite

        else:  # another version of the assembly found
            accession_v = accession_number_v[(accession_number_v.find(".")) + 1 : ]
            logger.warning(f"Multiple versions of {accession_v} found. Identifying the most recent")
            search_results.sort()
            existing_accession = existing_accession = search_results[-1]
            existing_v = existing_accession[(existing_accession.find(".")) + 1 : ]

            if accession_v > existing_v:
                accession_data = accession_data.pop(existing_v)
                accession_data[accession_number_v] = batch_result['DocumentSummarySet']['DocumentSummary'][index]['AssemblyName']

    for accession in tqdm(accession_data, desc=f"Downloading genomes for {term}"):
        for file_type in args.file_types:
            download_file(
                accession_number=accession,
                assembly_name=accession_data[accession],
                file_type=file_type,
                args=args,
            )
    return


def entrez_retry(entrez_func, *func_args, **func_kwargs):
    """Call to NCBI using Entrez.
    Maximum number of retries is 10, retry initated when network error encountered.
    :param logger: logger object
    :param retries: parser argument, maximum number of retries excepted if network error encountered
    :param entrez_func: function, call method to NCBI
    :param *func_args: tuple, arguments passed to Entrez function
    :param ** func_kwargs: dictionary, keyword arguments passed to Entrez function
    
    Returns a record.
    """
    logger = logging.getLogger(__name__)
    record, retries, tries = None, 10, 0

    while record is None and tries < retries:
        try:
            record = entrez_func(*func_args, **func_kwargs)

        except IOError:
            # log retry attempt
            if tries < retries:
                logger.warning(
                    f"Network error encountered during try no.{tries}.\nRetrying in 10s",
                    exc_info=1,
                )
                time.sleep(10)
            tries += 1

    if record is None:
        logger.error(
            "Network error encountered too many times. Exiting attempt to call to NCBI"
        )
        return

    return record


def download_file(
    accession_number, assembly_name, file_type, args, url_prefix="https://ftp.ncbi.nlm.nih.gov/genomes/all"
):
    """Download file.

    :param accession_number: str, accession number of genome
    :param assembly_name: str, name of assembly
    :param file_type: str, denotes in logger file type downloaded [accepted = 'protein.faa', 'genomic.fna']
    param args: parser arguments

    Return nothing.
    """
    logger = logging.getLogger(__name__)

    if args.database == 'genbank':   # retrieve GenBank not RefSeq
        gbk_accession = accession_number.replace("GCF_", "GCA_")
    else:  # retrieve RefSeq not GenBank
        gbk_accession = accession_number.replace("GCA_", "GCF_")

    file_name = f"{gbk_accession}_{assembly_name}.{file_type}"
    file_name = file_name.replace(" ","_")
    output_path = args.output_dir / file_name
    unzipped_path = Path(str(output_path).replace('.gz', ''))

    if output_path.exists():
        logger.warning(f"Output file {output_path} exists, not downloading")
        return
    if unzipped_path.exists():
        logger.warning(f"Output file {unzipped_path} exists, not downloading")
        return

    escape_characters = re.compile(r"[\s/,#\(\)]")
    escape_name = re.sub(escape_characters, "_", assembly_name)

    filestem = "_".join([gbk_accession, escape_name])

    url_parts = tuple(filestem.split("_", 2))
    # separate identifying numbers from version number, and split up into groups of 3 digits
    sub_directories = "/".join(
        [url_parts[1][i : i + 3] for i in range(0, len(url_parts[1].split(".")[0]), 3)]
    )

    genbank_url = f"{url_prefix}/{url_parts[0]}/{sub_directories}/{filestem}/{filestem}_{file_type}.gz"

    # try URL connection
    try:
        response = urlopen(genbank_url, timeout=args.timeout)
    except (HTTPError, URLError, timeout) as e:
        logger.error(
            f"Failed to download {file_type} for {accession_number}\nURL: {genbank_url}", exc_info=1,
        )
        return
    file_size = int(response.info().get("Content-length"))
    bsize = 1_048_576
    try:
        with open(output_path, "wb") as out_handle:
            # Using leave=False as this will be an internally-nested progress bar
            with tqdm(
                total=file_size,
                leave=False,
                desc=f"Downloading {accession_number} {file_type}",
            ) as pbar:
                while True:
                    buffer = response.read(bsize)
                    if not buffer:
                        break
                    pbar.update(len(buffer))
                    out_handle.write(buffer)
    except IOError:
        logger.error(f"Download failed for {accession_number}", exc_info=1)
        return

    return


if __name__ == "__main__":
    main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# (c) University of St Andrews 2022
# (c) University of Strathclyde 2022
# (c) James Hutton Institute 2022
# Author:
# Emma E. M. Hobbs

# Contact
# eemh1@st-andrews.ac.uk

# Emma E. M. Hobbs,
# Biomolecular Sciences Building,
# University of St Andrews,
# North Haugh Campus,
# St Andrews,
# KY16 9ST
# Scotland,
# UK

# The MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Functions to identify CAZy families that are always present together"""


import logging
import pandas as pd

from copy import copy
import matplotlib.pyplot as plt

import seaborn as sns
import upsetplot

from tqdm import tqdm


#
# Using a correlation matrix
#

def identify_cooccurring_fams_corrM(df, all_families, core_cazome=[], corrM_path=None, fill_value=2):
    """Build a correlation matrix of the CAZy fam frequencies to identify co-occurring families
    i.e. CAZy families that are always present together
    
    :param df: fam freq df, pandas df, columns are CAZy families, rows are genomes, cells 
        contain CAZy fam frequency
    :param all_families: set of all CAZy families to be analysed
    :param core_cazome: list of CAZy families in the core cazome, provide no list if you want to
        include the core CAZome
    :param corrM_path: path/str to write out the correlation matrix, if none, correlation matrix not
        written to a CSV file
    :param fill_value: int, value to replace Nan values in correlation matrix, if None - deletes columns
        and rows containing Nan values
        
    Return set of tuples, one tuple per group of co-occuring families
    and the filled correlation matrix
    """
    # convert CAZy fam freq df to binary presence/absence
    binary_fam_df = copy(df)
    
    for fam in tqdm(all_families, desc="Building binary fam freq df"):
        binary_fam_df.loc[binary_fam_df[fam] > 1, fam] = 1
        
    for fam in tqdm(all_families, desc="Delete absent families"):
        if 1 not in list(binary_fam_df[fam]):
            binary_fam_df.drop(fam, axis=1)
            
    for fam in core_cazome:
        binary_fam_df.drop(fam, axis=1)
        
    # correlation matrix
    fam_corr_M = binary_fam_df[all_families].corr()
    
    # write out a CSV file
    if corrM_path is not None:
        fam_corr_M.to_csv(corrM_path)
    
    if fill_value is None:
        fams_to_del = set()
        for fam in list(fam_corr_M.columns):
            for value in list(fam_corr_M[fam]):
                if str(value) == 'nan':
                    fams_to_del.add(fam)
                    
        for fam in fams_to_del:
            fam_corr_M.drop(fam, axis=1)
            fam_corr_M.drop(fam, axis=0)
                
    else:
        fam_corr_M_filled = fam_corr_M.fillna(fill_value)

    cooccurring_families = set()
    for fam in tqdm(all_families, desc="Identifying always co-occurring families"):
        if 1 not in list(fam_corr_M_filled[fam]):
            continue

        # find rows were value is 1
        value_1_rows = fam_corr_M_filled[fam_corr_M_filled[fam] == 1]
        if len(value_1_rows) == 1:
            continue

        fams = list(value_1_rows.index)
        fams.sort()
        fams = tuple(fams)
        cooccurring_families.add(fams)

    return cooccurring_families, fam_corr_M_filled


# 
# mannually parsing the df to identify co-occuring families
# including the number of genomes containing each group of CAZy families - needed for an upset plot
#

def calc_cooccuring_fam_freqs(df, all_families, exclude_core_cazome=False, core_cazome=[]):
    """Identify groups of CAZy families that are always present together, and count in 
    how many genomes the families are present together
    
    Initially calls another function to identify pairs of CAZy families that are always
    present together.
    
    Then identifies overlapping pairs of co-occurring CAZy families to identify grps 
    of co-occurring families, because if fam1 is always present with fam2, {fam1, fam2}
    and fam3 is always present with fam1 {fam1, fam3} then fam2 and fam3 must always
    be present together because both are always present with fam1.
    
    :param df: fam freq df, pandas df, columns are CAZy families, rows are genomes, cells 
        contain CAZy fam frequency
    :param all_families: set of all CAZy families to be analysed
    :param exclude_core_cazome: whether to exlude the core cazome, default: False - 
        include the core CAZome
    :param core_cazome: list of core CAZome families if to be excluded
        
    Return dict {grp_num: {'fams': {co-occurring fams}, 'freqs': {num of genomes}}
    - returns set of frequencies in case different numbers are produced for each inital pair 
    of co-occurring families
    """
    cooccuring_fams_dict = identify_cooccurring_fam_pairs(df, all_families, exclude_core_cazome=exclude_core_cazome)
    
    cooccurring_groups = {}
    grp_num = 0
    
    for cofams in tqdm(cooccuring_fams_dict, desc='Combining pairs of co-occurring families'):
        fams = cooccuring_fams_dict[cofams]['fams']
        
        added = False
        for grp in cooccurring_groups:
            if fams[0] in cooccurring_groups[grp]['fams']:
                cooccurring_groups[grp]['fams'].add(fams[1])
                cooccurring_groups[grp]['freqs'].add(cooccuring_fams_dict[cofams]['freq'])
                added = True

        if added is False:
            for grp in cooccurring_groups:
                if fams[1] in cooccurring_groups[grp]['fams']:
                    cooccurring_groups[grp]['fams'].add(fams[0])
                    cooccurring_groups[grp]['freqs'].add(cooccuring_fams_dict[cofams]['freq'])
                    added = True

        if added is False:
            cooccurring_groups[grp_num] = {
                'fams': {fams[0], fams[1]},
                'freqs': {cooccuring_fams_dict[cofams]['freq']}
            }
            grp_num+=1

    for grp in cooccurring_groups:
        if len(cooccurring_groups[grp]['freqs']) > 1:
            print(f':WARNING: differing freqs found for grp: {cooccurring_groups[grp]}')
    
    return cooccurring_groups
    
    
def identify_cooccurring_fam_pairs(df, all_families, exclude_core_cazome=False, core_cazome=[]):
    """Identify pairs of CAZy families that are always present together in the same genome
    
    :param df: fam freq df, pandas df, columns are CAZy families, rows are genomes, cells 
        contain CAZy fam frequency
    :param all_families: set of all CAZy families to be analysed
    :param exclude_core_cazome: whether to exlude the core cazome, default: False - 
        include the core CAZome
        
    Return dict {str(tuple(fams)): {'fams': tuple(fams), 'freq': int(num of genomes)}}
    """
    logger = logging.getLogger(__name__)
    cooccuring_fams_dict = {}  # {str(tuple(fams)): {'fams': tuple(fams), 'freq': int(num of genomes)}}
    
    for current_fam in tqdm(all_families, desc="Identifying pairs of co-occurring families"):
        other_fams = copy(all_families)
        other_fams.remove(current_fam)

        current_fam_freqs = list(df[current_fam])

        bin_current_freqs = []
        for value in current_fam_freqs:
            if value > 1:
                bin_current_freqs.append(1)
            else:
                bin_current_freqs.append(value)

        if 1 not in bin_current_freqs:
            continue  # fam not in this data set

        for other_fam in other_fams:
            other_fam_freqs = list(df[other_fam])

            bin_other_freqs = []
            for value in other_fam_freqs:
                if value > 1:
                    bin_other_freqs.append(1)
                else:
                    bin_other_freqs.append(value)

            if 1 not in bin_other_freqs:
                continue  # other_fam not in this data set

            # test that when one is present the other is
            cooccurring = True
            for i in range(len(df)):
                if (current_fam_freqs[i] != 0) and (other_fam_freqs[i] == 0):
                    cooccurring = False
                    break
                elif (current_fam_freqs[i] == 0) and (other_fam_freqs[i] != 0):
                    cooccurring = False
                    break

            if cooccurring is False:
                continue  # on to the next other_fam

            # count_genomes
            freq = 0
            for i in range(len(bin_current_freqs)):
                if (bin_current_freqs[i] == 1) and (bin_other_freqs[i] == 1):
                    freq += 1

            if freq == 0:
                continue  # on to the next other_fam

            # check if one family is always absend when the other family is absent
            if exclude_core_cazome:
                if (current_fam in core_cazome) and (other_fam in core_cazome):
                    continue  # core cazome

            families = [current_fam, other_fam]
            families.sort()
            families = tuple(families)
            try:
                cooccuring_fams_dict[str(families)]
                if cooccuring_fams_dict[str(families)]['freq'] != freq:
                    logger.warning(
                        f'Found {current_fam} and {other_fam} previously with '
                        f'freq {cooccuring_fams_dict[str(families)]["freq"]}\n'
                        f'but calculated new freq of {freq}.\n'
                        'Using original calculated freq'
                    )
            except KeyError:
                cooccuring_fams_dict[str(families)] = {'fams': families, 'freq': freq}

    return cooccuring_fams_dict


#
# Build upset plots
#


def add_to_upsetplot_membership(upsetplot_membership, cooccurring_fams_dict):
    """Add co-occurring families to upsetplot membership data
    
    :param upsetplot_membership: list of lists, one nested list per instance of co-occurring families group
    :param cooccurring_fams_dict, dict {grp_num: {'fams': {families}: 'freqs': {ints(num of genomes)}}}
    
    Return updated list of upsetplot membership
    """
    for grp_num in cooccurring_fams_dict:
        families = list(cooccurring_fams_dict[grp_num]['fams'])
        freq = list(cooccurring_fams_dict[grp_num]['freqs'])[0]
        for i in range(freq):
            upsetplot_membership.append(families)
    return upsetplot_membership


def build_upsetplot(
    upsetplot_membership,
    file_path=None,
    file_format='svg',
    sort_by='degree',
    sort_categories_by='cardinality',
):
    """Use the upsetplot package to build an upsetplot of co-occurring families
    
    :param upsetplot_membership: list of lists, one nested list per instance of co-occurring families group
    :param file_path, str/Path, path to write out figure. If none, file is not written out
    :param file_format: str, format to write out file, e.g. svg or png, default, svg
    :param sort_by: str, method to sort subsets 
        From Upsetplot:
            sort_by : {'cardinality', 'degree', '-cardinality', '-degree',
                    'input', '-input'}
                If 'cardinality', subset are listed from largest to smallest.
                If 'degree', they are listed in order of the number of categories
                intersected. If 'input', the order they appear in the data input is
                used.
                Prefix with '-' to reverse the ordering.

                Note this affects ``subset_sizes`` but not ``data``.
    :param sort_categories_by: str, 
        From UpsetPlot:
            sort_categories_by : {'cardinality', '-cardinality', 'input', '-input'}
                Whether to sort the categories by total cardinality, or leave them
                in the input data's provided order (order of index levels).
                Prefix with '-' to reverse the ordering.
    
    Return upsetplot
    """
    upset_data = upsetplot.from_memberships(upsetplot_membership)
    coocurring_upset_plot = upsetplot.UpSet(
        upset_data,
        subset_size='sum',
        sort_categories_by=sort_categories_by,
        sort_by=sort_by,
    )
    coocurring_upset_plot.plot();
    
    if file_path is not None:
        plt.savefig(file_path, format=file_format)

    return coocurring_upset_plot


def get_upsetplot_grps(upsetplot_membership):
    """Retrieve the groups of CAZy families in the upset plot, in the order they are presented in the plot.
    
    :param upsetplot_membership: list of lists, membership data used to build the upset plot
    
    Return list of lists, one nested list per group of co-occurring CAZy families
    """
    upset_data = upsetplot.from_memberships(upsetplot_membership)
    
    upsetplot_grp_data = upsetplot.query(
        upset_data, 
        subset_size='count',
        sort_by='degree',
    )
    
    # extract the presence/absence data of each CAZy family per group of families in the upsetplot
    # convert series to a df, personally easier to handle and visualise
    upset_plot_df = pd.DataFrame(upsetplot_grp_data.subset_sizes)
    # upsetplot_df contains one row per group in the upset plot
    # going down the df, shows each group left to right on the upset plot
    
    upset_plot_groups = []

    upsetplot_df_fams = list(upset_plot_df.index.names)
    
    for ri in tqdm(range(len(upset_plot_df))):
        # the name of each row is a tuple of boolean values, one value per fam
        # marking if fam in group or not
        grp = []

        for i in range(len(upsetplot_df_fams)):
            if upset_plot_df.iloc[ri].name[i]: # is True
                grp.append(upsetplot_df_fams[i])

        upset_plot_groups.append(grp)
        
    return upset_plot_groups


def add_upsetplot_grp_freqs(
    upset_plt_groups,
    cooccurring_grp_freq_data,
    cooccurring_fam_dict,
    grp,
    grp_sep=False,
    grp_order=None,
    include_none=False,
):
    """Add data on the incidence of co-occurring grps of CAZy families from the
    cooccurring_fam_dict to cooccurring_grp_freq_data
    
    :param upset_plt_groups: list of lists, one nested list per grp of co-occurring CAZy families
        grps listed in same order as present in the upsetplot
    :param cooccurring_grp_freq_data: list of lists, one nested list per 
        pair of 'grp' and grp of co-occurring CAZy families
    :param cooccurring_fam_dict: dict, {grp_num: {'fams': {families}, 'freqs': {freqs/incidences}}}
    :param grp: str, name of grp to be added to cooccurring_grp_freq_data, e.g. the name of the genus
    :param grp_sep: bool, does the cooccurring_fam_dict contain data separated into grps, e.g. by genus
        {grp(e.g. genus): {grp_num: {'fams': {families}, 'freqs': {freqs/incidences}}}}
    :param grp_order: list of grp names, order to list through grps if grp_sep is True. If None, uses
        order groups are listed in cooccurring_fam_dict
    :param include_none: bool, if True, if a grp of fams is not in the cooccurring_fam_dict, leaves the 
        freq as None. If false, the grp of fams is not added to upset_plt_groups
    
    Return cooccurring_grp_freq_data
    """

    for fam_grp in tqdm(upset_plt_groups, desc="Compiling co-occurring families incidence data"):

        if grp_sep:
            if grp_order is None:
                for grp_name in cooccurring_fam_dict:
                    added = False
                    for grp_num in cooccurring_fam_dict[grp_name]:
                        if cooccurring_fam_dict[grp_name][grp_num]['fams'] == set(fam_grp):
                            cooccurring_grp_freq_data.append(
                                [
                                    "+".join(fam_grp),
                                    grp_name,
                                    list(cooccurring_fam_dict[grp_name][grp_num]['freqs'])[0],
                                ]
                            )
                            added = True
                            
                    if (include_none) and (added is False):
                        cooccurring_grp_freq_data.append(
                            [
                                "+".join(fam_grp),
                                grp_name,
                                None,
                            ]
                        )
                        
            else:
                for grp_name in grp_order:
                    added = False
                    for grp_num in cooccurring_fam_dict[grp_name]:
                        if cooccurring_fam_dict[grp_name][grp_num]['fams'] == set(fam_grp):
                            cooccurring_grp_freq_data.append(
                                [
                                    "+".join(fam_grp),
                                    grp_name,
                                    list(cooccurring_fam_dict[grp_name][grp_num]['freqs'])[0],
                                ]
                            )
                            added = True
                            
                    if (include_none) and (added is False):
                        cooccurring_grp_freq_data.append(
                            [
                                "+".join(fam_grp),
                                grp_name,
                                None,
                            ]
                        )

        else:
            # get the data for the relevant group
            added = False
            for grp_num in cooccurring_fam_dict:
                if cooccurring_fam_dict[grp_num]['fams'] == set(fam_grp):
                    cooccurring_grp_freq_data.append(
                        [
                            "+".join(fam_grp),
                            grp,
                            list(cooccurring_fam_dict[grp_num]['freqs'])[0],
                        ]
                    )
                    added = True
            
            if (include_none) and (added is False):
                cooccurring_grp_freq_data.append(
                    [
                        "+".join(fam_grp),
                        grp,
                        None,
                    ]
                        )
                
    return cooccurring_grp_freq_data


def build_upsetplot_matrix(cooccurring_grp_freq_data, grp, file_path=None):
    """Build matrix of grp of CAZy families, grp of interest name (e.g. genus) and incidence 
    (i.e. the number of genomes that the grp of CAZy families appeared in)
    
    :param cooccurring_grp_freq_data: list of lists, one nested list per row in the df
    :param grp: str, name of grouping, i.e. the method used to group the genomes,
        .e.g. 'Genus', or 'Species'
    :param file_path: str/Path, path to write out CSV file. If none, the file is not 
        written to file
        
    Return df
    """
    df = pd.DataFrame(cooccurring_grp_freq_data, columns=[
        'Families',
        grp,
        'Incidence',
    ])
    
    if file_path is not None:
        df.to_csv(file_path)
    
    return df
